package net.rten.palmve.data.pool;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;

import android.util.Log;

import androidx.arch.core.executor.testing.InstantTaskExecutorRule;
import androidx.lifecycle.LiveData;

import net.rten.palmve.repository.PoolRepo;
import net.rten.palmve.utilities.LiveDataTestUtil;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.inject.Inject;

import dagger.hilt.android.testing.HiltAndroidRule;
import dagger.hilt.android.testing.HiltAndroidTest;

@HiltAndroidTest
public class TestPoolDatabase {
    private static final String TAG = "TestPoolDatabase";
    @Rule
    public final HiltAndroidRule _hiltRule = new HiltAndroidRule(this);
    @Rule
    public InstantTaskExecutorRule _executor = new InstantTaskExecutorRule();
    @Inject
    protected PoolRepo _pools;

    @Before
    public void before() {
        _hiltRule.inject();
    }

    @Test
    public void testElement() {
        Element e = getPoolDb().elementDao().getElement();
        assertEquals("EXTRA CLASS", e.name);
        assertEquals("July 1, 2020 throuh June 30, 2024", e.dateField);

        LiveData<ElementView> elementLive = getPoolDb().elementDao().getElementLiveData();
        ElementView elementView = LiveDataTestUtil.getOrAwaitValue(elementLive);
        assertEquals("EXTRA CLASS", elementView.designation);
        assertEquals("July 1, 2020 throuh June 30, 2024", elementView.description);
    }

    @Test
    public void testQuestionText() {
        Log.i("TestPoolDatabase", "something");
        Question q = getPoolDb().questionDao().findQuestionByIdentifiers("E9", "H", "07");
        String s = getPoolDb().questionDao().getTextForQuestionId(q.id);
        assertEquals(s, "Why is RF attenuation used when direction-finding?");
    }

    @Test
    public void testQuestionsAnswersAndSubElements() {
        List<SubElementView> se = getPoolDb().subElementDao().subElementsInElement(4);
        assertEquals(10, se.size());

        List<QuestionView> qvs = getPoolDb().questionDao().questions();
        assertEquals(622, qvs.size());

        List<AnswerView> av = getPoolDb().questionDao().getAnswers("E9", "H", "07");
        assertEquals(4, av.size());

        List<QuestionGroupView> qg = getPoolDb().questionGroupsDao().questionGroupsInSubElement(9);
        assertEquals(8, qg.size());
    }

    @Test
    public void testSearching() {
        List<QuestionView> sq = getPoolDb().questionDao().searchQuestions("mosfet");
        assertEquals(2, sq.size());

        List<AnswerView> sa = getPoolDb().questionDao().searchAnswers("transistor");
        assertEquals(6, sa.size());

        LiveData<List<QuestionViewForAnswerSearch>> lsqa = getPoolDb().questionDao().searchQuestionAnswersLiveData("transistor");
        List<QuestionViewForAnswerSearch> sqa = LiveDataTestUtil.getOrAwaitValue(lsqa);
        assertEquals(sa.size(), sqa.size());
    }

    @Test
    public void testGenerateATest() {
        List<QuestionView> questions = getPoolDb().testGenerationDao().generateTest();

        // Extra tests should have 50 questions. We should only see one per question group and
        // none should be withdrawn.
        assertEquals(50, questions.size());
        HashMap<Long, Long> questionGroups = new HashMap<>();
        Log.i(TAG, "Questions in new test:");
        for (QuestionView q : questions) {
            Log.i(TAG, " .. " + q.description);
            assertNotNull(q.questionGroupId);
            //noinspection ConstantConditions
            questionGroups.put(q.questionGroupId,
                    questionGroups.getOrDefault(q.questionGroupId, 0L) + 1);
            assertFalse(q.withdrawn);
        }
        for (Map.Entry<Long, Long> entry : questionGroups.entrySet()) {
            assertEquals(new Long(1), entry.getValue());
        }
    }

    private PoolDatabase getPoolDb() {
        return _pools.getLocalPools().get(0).getRoomDb();
    }
}
