package net.rten.palmve.data.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;

import androidx.arch.core.executor.testing.InstantTaskExecutorRule;
import androidx.lifecycle.LiveData;

import net.rten.palmve.TestEnvironment;
import net.rten.palmve.data.pool.Element;
import net.rten.palmve.data.pool.PoolDatabase;
import net.rten.palmve.data.pool.QuestionView;
import net.rten.palmve.repository.PoolRepo;
import net.rten.palmve.repository.TestHistoryRepo;
import net.rten.palmve.utilities.LiveDataTestUtil;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;

import java.io.IOException;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.GregorianCalendar;
import java.util.List;

import javax.inject.Inject;

import dagger.hilt.android.testing.HiltAndroidRule;
import dagger.hilt.android.testing.HiltAndroidTest;

@HiltAndroidTest
public class TestTestDatabase {
    @Rule
    public InstantTaskExecutorRule _executor = new InstantTaskExecutorRule();
    @Rule
    public final HiltAndroidRule _hiltRule = new HiltAndroidRule(this);
    @Inject
    public TestHistoryRepo _tests;
    @Inject
    public PoolRepo _pools;
    @Inject
    public TestEnvironment _testEnv;

    @Before
    public void before() throws IOException {
        // Start with a fresh test history DB for every test.
        _hiltRule.inject();
        _tests.close();
        _testEnv.reset();
        _tests.reset();
    }

    @Test
    public void testTestList() {
        LiveData<List<TestHistory>> testsLiveData = getTestDb().testDao().getAllTests();
        List<TestHistory> tests = LiveDataTestUtil.getOrAwaitValue(testsLiveData);
        assertEquals(4, tests.size());

        // Simply makes sure the Calendar->Long conversion stuff works.
        for (TestHistory test : tests) {
            GregorianCalendar g = (GregorianCalendar) test.dateStarted;
            LocalDate date = g.toZonedDateTime().toLocalDate();
            assertEquals("2021-05-05", date.format(DateTimeFormatter.ISO_DATE));
            assertEquals(TestState.FAILED, test.testState);
        }
    }

    @Test
    public void testGetQuestionsInTest() {
        LiveData<List<TestQuestions>> questionsLiveData = getTestDb().testDao().getQuestionsInTestLiveData(1);
        List<TestQuestions> questions = LiveDataTestUtil.getOrAwaitValue(questionsLiveData);
        assertEquals(35, questions.size());
        int expectedIndex = 0;
        for (TestQuestions q : questions) {
            assertEquals(expectedIndex++, q.indexOfQuestionInTest.intValue());
        }
    }

    @Test
    public void testQuestionGroupStats() {
        // The GENERAL test history has two tests, so we should see some entries with two questions.
        LiveData<List<StatsView>> correctLive =
                getTestDb().statsDao().getQuestionGroupStatsWithCorrectAnswers(
                        "GENERAL CLASS", "July 1, 2019 through June 30, 2023", -1);
        List<StatsView> correct = LiveDataTestUtil.getOrAwaitValue(correctLive);
        assertEquals(17, correct.size());
        assertEquals("G2A??", correct.get(0).designation);
        assertNull(correct.get(0).questionId);
        assertEquals(6, correct.get(0).questionGroupId.intValue());
        assertEquals(2, correct.get(0).questionCount.intValue());
        assertEquals(true, correct.get(0).answersWereCorrect);

        LiveData<List<StatsView>> incorrectLive =
                getTestDb().statsDao().getQuestionGroupStatsWithIncorrectAnswers(
                        "GENERAL CLASS", "July 1, 2019 through June 30, 2023", -1);
        List<StatsView> incorrect = LiveDataTestUtil.getOrAwaitValue(incorrectLive);
        assertEquals(34, incorrect.size());
        assertEquals("G0B??", incorrect.get(0).designation);
        assertNull(incorrect.get(0).questionId);
        assertEquals(35, incorrect.get(0).questionGroupId.intValue());
        assertEquals(2, incorrect.get(0).questionCount.intValue());
        assertEquals(false, incorrect.get(0).answersWereCorrect);
    }

    @Test
    public void testQuestionStats() {
        LiveData<List<StatsView>> correctLive =
                getTestDb().statsDao().getQuestionStatsWithCorrectAnswers(
                        "GENERAL CLASS", "July 1, 2019 through June 30, 2023", -1);
        List<StatsView> correct = LiveDataTestUtil.getOrAwaitValue(correctLive);
        assertEquals(18, correct.size());
        assertEquals("G0A08", correct.get(0).designation);
        assertEquals(437, correct.get(0).questionId.intValue());
        assertEquals(true, correct.get(0).answersWereCorrect);

        LiveData<List<StatsView>> incorrectLive =
                getTestDb().statsDao().getQuestionStatsWithIncorrectAnswers(
                        "GENERAL CLASS", "July 1, 2019 through June 30, 2023", -1);
        List<StatsView> incorrect = LiveDataTestUtil.getOrAwaitValue(incorrectLive);
        assertEquals(50, incorrect.size());
        assertEquals("G1D09", incorrect.get(0).designation);
        assertEquals(51, incorrect.get(0).questionId.intValue());
        assertEquals(false, incorrect.get(0).answersWereCorrect);
    }

    @Test
    public void testTestLists() {
        LiveData<List<TestCount>> countsLive = getTestDb().testDao().getTestCounts(
                "GENERAL CLASS", "July 1, 2019 through June 30, 2023");
        List<TestCount> counts = LiveDataTestUtil.getOrAwaitValue(countsLive);
        assertEquals(1, counts.size());
        assertEquals(2, counts.get(0).testCount.intValue());
        assertEquals(TestState.FAILED, counts.get(0).testState);
    }

    @Test
    public void testActiveTests() {
        assertNull(getTestDb().activeTestDao().getActiveTestOrNull(
                "GENERAL CLASS", "July 1, 2019 through June 30, 2023"
        ));
    }

    @Test
    public void testGenerateATest() {
        // We need a pool database to generate a test.
        Element element = getPoolDb().elementDao().getElement();
        List<QuestionView> questions = getPoolDb().testGenerationDao().generateTest();

        // Now we can add the test to the test DB.
        final long testId = getTestDb().activeTestDao().createTest(getTestDb().testDao(), element, questions);

        // Read it back and see what we got.
        LiveData<List<TestQuestions>> questionsInNewTestLive =
                getTestDb().testDao().getQuestionsInTestLiveData(testId);
        List<TestQuestions> questionsInNewTest = LiveDataTestUtil.getOrAwaitValue(questionsInNewTestLive);
        assertEquals(questionsInNewTest.size(), questions.size());
        for (int i = 0; i < questions.size(); ++i) {
            assertEquals(questions.get(i).designation, questionsInNewTest.get(i).questionDesignation);
        }

        LiveData<TestHistory> testLive = getTestDb().testDao().getTestInfoByIdLiveData(testId);
        TestHistory test = LiveDataTestUtil.getOrAwaitValue(testLive);
        assertEquals(test.id, new Long(testId));
    }

    private TestDatabase getTestDb() {
        return _tests.getDb();
    }

    private PoolDatabase getPoolDb() {
        return _pools.getLocalPools().get(0).getRoomDb();
    }
}
