/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve;

import android.content.Context;
import android.database.DataSetObserver;
import android.view.View;
import android.view.ViewGroup;
import android.widget.ListAdapter;
import android.widget.ListView;

import net.rten.palmve.repository.PoolDirectoryObserver;
import net.rten.palmve.repository.PoolRepo;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Adapts the list of pools in a pool directory to a list view.
 */
public class PoolListAdapter implements ListAdapter {
    final private PoolRepoObserver _repoEvents = new PoolRepoObserver();
    final private List<ItemWrapper> _items = new ArrayList<>();
    final private PoolRepo _poolRepo;
    final private Context _context;
    private DataSetObserver _observer;
    private boolean _showBadNetworkAlert;
    private String _badNetworkMessage;

    public PoolListAdapter(Context context, PoolRepo poolRepoService) {
        _poolRepo = poolRepoService;
        _poolRepo.addObserver(_repoEvents);
        _context = context;
        _showBadNetworkAlert = false;
        generateDisplayPools();
    }

    /**
     * Figure out which pools we're going to display, and put them in our
     * internal pool display list. If we're displaying both types, We have to
     * merge the local and remote pools into a single list of what to display.
     * We shouldn't display a remote pool when we have a local copy.
     */
    private void generateDisplayPools() {
        // a comparator for pools. uses element number
        @SuppressWarnings("ComparatorCombinators") Comparator<Pool> poolComparator = (p1, p2) -> p1.getElementNr() - p2.getElementNr();

        // reset the display pools list
        _items.clear();

        // local pools
        final ArrayList<LocalPool> masterLocalPools = _poolRepo.getLocalPools();
        if (masterLocalPools.size() != 0) {
            // sort the pools into a reasonable order first
            ArrayList<LocalPool> sortedLocalPools = new ArrayList<>(
                    masterLocalPools.size());
            sortedLocalPools.addAll(masterLocalPools);
            Collections.sort(sortedLocalPools, poolComparator);

            // add the header and items
            if (sortedLocalPools.size() != 0) {
                _items.add(new ItemWrapper(_context.getString(R.string.PoolListHeader_LocalPools),
                        ItemWrapper.TYPE_HEADER));
                for (LocalPool localPool : sortedLocalPools) {
                    _items.add(new ItemWrapper(localPool, ItemWrapper.TYPE_POOL));
                }
            }
        }

        // remote pools
        final ArrayList<RemotePool> masterRemotePools = _poolRepo.getRemotePools();
        if (masterRemotePools != null) {
            ArrayList<RemotePool> sortedRemotePools = new ArrayList<>(
                    masterRemotePools.size());

            // gather and sort applicable remote pools
            for (RemotePool remotePool : _poolRepo.getRemotePools()) {
                if (remotePool.getAssociatedLocalPool() == null) {
                    sortedRemotePools.add(remotePool);
                }
            }
            Collections.sort(sortedRemotePools, poolComparator);

            // now add header and items
            if (sortedRemotePools.size() != 0) {
                _items.add(new ItemWrapper(_context.getString(R.string.PoolListHeader_RemotePools),
                        ItemWrapper.TYPE_HEADER));

                for (RemotePool remotePool : sortedRemotePools) {
                    _items.add(new ItemWrapper(remotePool, ItemWrapper.TYPE_POOL));
                }
            }
        }

        // if we have no items... well, it means we must download some to be
        // useful
        if (_showBadNetworkAlert && _items.isEmpty()) {
            String line1, line2;

            line1 = _context.getString(R.string.PoolListAdapter_NoPoolsAlertL1);
            if (_badNetworkMessage == null) {
                line2 = _context.getString(R.string.PoolListAdapter_NoPoolsAlertL2);
            } else {
                line2 = String.format(_context.getString(
                        R.string.PoolListAdapter_NoPoolsAlertL2WErr), _badNetworkMessage);
            }

            _items.add(new ItemWrapper(new String[]{line1, line2}, ItemWrapper.TYPE_ALERT));
        }

        // alert the observer
        if (_observer != null) {
            _observer.onChanged();
        }
    }

    // ListAdapter:
    @Override
    public boolean areAllItemsEnabled() {
        return false;
    }

    @Override
    public boolean isEnabled(int position) {
        return _items.get(position).getObjectType() != ItemWrapper.TYPE_HEADER;
    }

    // Adapter:
    @Override
    public int getCount() {
        return _items.size();
    }

    @Override
    public Object getItem(int position) {
        return _items.get(position);
    }

    @Override
    public long getItemId(int position) {
        return position;
    }

    @Override
    public int getItemViewType(int position) {
        return _items.get(position).getObjectType();
    }

    @Override
    public View getView(int position, View convertView, ViewGroup parent) {
        ItemWrapper item = _items.get(position);
        View returnValue = null;
        Class<?> expectedViewClass = null;

        switch (item.getObjectType()) {
            case ItemWrapper.TYPE_POOL:
                expectedViewClass = PoolListQuestionPool.class;
                break;
            case ItemWrapper.TYPE_HEADER:
                expectedViewClass = PoolListSeparator.class;
                break;
            case ItemWrapper.TYPE_ALERT:
                expectedViewClass = PoolListAlert.class;
                break;
        }

        if (expectedViewClass != null) {
            try {
                PoolListItem listItem;
                if (convertView == null
                        || !expectedViewClass.isInstance(convertView.getTag())) {
                    listItem = (PoolListItem) expectedViewClass.newInstance();
                    listItem.setContext(_context);
                } else {
                    listItem = (PoolListItem) convertView.getTag();
                }
                listItem.setListObject(item.getObject());
                returnValue = listItem.getView();
            } catch (IllegalAccessException | InstantiationException e) {
                e.printStackTrace();
            }
        }

        return returnValue;
    }

    @Override
    public int getViewTypeCount() {
        return ItemWrapper.TYPE_COUNT;
    }

    @Override
    public boolean hasStableIds() {
        return false;
    }

    @Override
    public boolean isEmpty() {
        return getCount() == 0;
    }

    @Override
    public void registerDataSetObserver(DataSetObserver observer) {
        _observer = observer;
    }

    @Override
    public void unregisterDataSetObserver(DataSetObserver observer) {
        _observer = null;
    }

    /**
     * Close out any of our data, to help GC do its job. This is really
     * unfortunate, but to prevent memory leaks it needs doing. You need to pass
     * the list view that was associated with this item, so that we can clear
     * out its item views.
     */
    public void close(ListView view) {
        _poolRepo.removeObserver(_repoEvents);
        _items.clear();

        ArrayList<View> views = new ArrayList<>();
        view.reclaimViews(views);
        for (View itemView : views) {
            PoolListItem pli = (PoolListItem) itemView.getTag();
            if (pli != null) {
                pli.setListObject(null);
            }
        }
    }

    /**
     * Wraps our items, since we can contain headers, pool objects, and alerts
     */
    public final static class ItemWrapper {
        public final static int TYPE_POOL = 0;
        public final static int TYPE_ALERT = 2;
        final static int TYPE_HEADER = 1;
        final static int TYPE_COUNT = TYPE_ALERT + 1;

        private final Object _object;
        private final int _objectType;

        ItemWrapper(Object object, int objectType) {
            _objectType = objectType;
            _object = object;
        }

        public Object getObject() {
            return _object;
        }

        public int getObjectType() {
            return _objectType;
        }
    }

    private class PoolRepoObserver implements PoolDirectoryObserver {
        /**
         * The pool list has changed. Update our display list.
         */
        @Override
        public void onPoolDirectoryChanged(PoolRepo directory) {
            generateDisplayPools();
        }

        /**
         * We might have an alert in the display list. Remove it if so, so that it
         * doesn't hide behind any active progress dialog.
         */
        @Override
        public void onDownloadingRemotePoolDataBeginning(PoolRepo directory) {
            generateDisplayPools();
        }

        /**
         * We do nothing. We don't care about this event.
         */
        @Override
        public void onDownloadingRemotePoolDataCompleted(PoolRepo directory) {
        }

        /**
         * Refresh. We may need to show an alert.
         */
        @Override
        public void onDownloadingRemotePoolDataFailed(PoolRepo directory,
                                                      String message) {
            _showBadNetworkAlert = true;
            _badNetworkMessage = message;
            generateDisplayPools();
            _showBadNetworkAlert = false;
            _badNetworkMessage = null;
        }
    }
}
