/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve;

import android.content.Context;
import android.text.format.Formatter;
import android.view.View;
import android.widget.ImageView;
import android.widget.ProgressBar;
import android.widget.RelativeLayout;
import android.widget.TextView;

import net.rten.palmve.LocalPool.Status;

import java.text.DateFormat;

/**
 * An item in the pool list. Handles updating the UI based on pool changes.
 *
 * @author ballbach
 */
public class PoolListQuestionPool implements PoolListItem, RemotePoolEvents,
        LocalPoolEvents {
    private RelativeLayout _layout;
    private TextView _line1, _line2, _tag, _line3;
    private ProgressBar _progressBar;
    private ImageView _icon;
    private int _currentIconResource;
    private Pool _pool;

    public PoolListQuestionPool() {
    }

    public void setContext(Context context) {
        _layout = (RelativeLayout) View.inflate(context, R.layout.pool_item, null);
        _tag = _layout.findViewById(R.id.poolItemSecondLineTag);
        _line1 = _layout.findViewById(R.id.poolItemFirstLine);
        _line2 = _layout.findViewById(R.id.poolItemSecondLine);
        _line3 = _layout.findViewById(R.id.poolItemThirdLine);
        _progressBar = _layout.findViewById(R.id.poolItemProgressBar);
        _icon = _layout.findViewById(R.id.poolItemIcon);
        _layout.setTag(this);
    }

    public View getView() {
        return _layout;
    }

    /**
     * Bind the UI element to a pool.
     */
    public void setListObject(Object object) {
        Pool pool = (Pool) object;

        // if we're bound to a pool, we might need to reset
        if (_pool != null) {
            if (_pool instanceof RemotePool) {
                RemotePool rp = (RemotePool) _pool;
                rp.removeEventHandler(this);
            } else if (_pool instanceof LocalPool) {
                LocalPool lp = (LocalPool) _pool;
                RemotePool rp = lp.getAssociatedRemotePool();
                if (rp != null) {
                    rp.removeEventHandler(this);
                }
                lp.removeEventHandler(this);
            }
        }
        _pool = pool;
        _currentIconResource = -1;

        if (_pool != null) {
            // update the UI
            updateUI();

            // register for events
            if (_pool instanceof RemotePool) {
                RemotePool rp = (RemotePool) _pool;
                rp.addEventHandler(this);
            } else if (_pool instanceof LocalPool) {
                LocalPool lp = (LocalPool) _pool;
                RemotePool rp = lp.getAssociatedRemotePool();
                lp.addEventHandler(this);
                if (rp != null) {
                    // if the lp has an rp, we register for its events, too,
                    // because the user might start downloading a new version of
                    // it.
                    rp.addEventHandler(this);
                }
            }
        }
    }

    /**
     * Update the UI to reflect the state of the pool.
     */
    private void updateUI() {
        final Context context = _layout.getContext();

        Pool pool = _pool;
        LocalPool localPool = null;
        RemotePool remotePool;
        int iconResource;
        String tagText = "";

        if (pool instanceof RemotePool) {
            remotePool = (RemotePool) pool;
        } else {
            localPool = (LocalPool) pool;
            remotePool = localPool.getAssociatedRemotePool();
        }

        // format some text for the date field line
        String dateText = pool.getDateField();
        LocalPool.EffectiveDateRange dateRange = LocalPool
                .getEffectiveDateRange(dateText);
        int dateTextColor = _line1.getCurrentTextColor();
        if (dateRange.from != null && dateRange.to != null) {
            int dateStatus = pool.getDateStatus(), strResID;

            switch (dateStatus) {
                case LocalPool.DATE_STATUS_EXPIRED:
                    strResID = R.string.PoolListQP_Expired;
                    dateTextColor = 0xffff3333;
                    break;

                case LocalPool.DATE_STATUS_FUTURE:
                    strResID = R.string.PoolListQP_Future;
                    dateTextColor = 0xffcccc33;
                    break;

                case LocalPool.DATE_STATUS_ACTIVE:
                default:
                    strResID = R.string.PoolListQP_Active;
                    break;
            }

            DateFormat dateFormat = DateFormat.getDateInstance(DateFormat.LONG);
            dateFormat.setCalendar(dateRange.from);
            dateText = context.getString(strResID,
                    dateFormat.format(dateRange.from.getTime()),
                    dateFormat.format(dateRange.to.getTime()));
        }

        String sizeText;
        if (pool.getSize() <= 0) {
            sizeText = context.getString(R.string.PoolList_SizeUnknown);
        } else {
            String formattedSize = Formatter.formatFileSize(context, pool.getSize());
            if (localPool != null) {
                sizeText = context.getString(R.string.PoolList_SizeLocalPool, formattedSize);
            } else {
                sizeText = context.getString(R.string.PoolList_SizeRemotePool, formattedSize);
            }
        }

        // set our basic text lines
        _line1.setText(pool.getName());
        _line2.setText(dateText);
        _line2.setTextColor(dateTextColor);
        _line3.setText(sizeText);

        // show or hide the progress bar as necessary
        if (remotePool != null
                && remotePool.getStatus() == RemotePool.Status.rpDownloading) {
            if (_line2.getVisibility() != View.GONE) {
                _line2.setVisibility(View.GONE);
                _tag.setVisibility(View.GONE);
            }
            if (_progressBar.getVisibility() != View.VISIBLE) {
                _progressBar.setVisibility(View.VISIBLE);
            }
            _progressBar.setProgress(remotePool.getDownloadProgress());
        } else {
            if (_line2.getVisibility() != View.VISIBLE) {
                _line2.setVisibility(View.VISIBLE);
                _tag.setVisibility(View.VISIBLE);
            }
            if (_progressBar.getVisibility() != View.GONE) {
                _progressBar.setVisibility(View.GONE);
            }
        }

        // if we have a download error, display it
        if (remotePool != null) {
            if (remotePool.getStatus() == RemotePool.Status.rpDownloadFailed) {
                _line2.setText(remotePool.getDownloadErrorString());
            }
        }

        // update the icon
        if (localPool != null) {
            if (localPool.getStatus() == LocalPool.Status.lpUpdateAvailable) {
                iconResource = R.drawable.ic_upgradable_pool;
                tagText = "Upgrade Now: ";
            } else {
                iconResource = R.drawable.icon;
            }
        } else {
            iconResource = R.drawable.ic_downloadable_pool;
        }
        if (_currentIconResource != iconResource) {
            _icon.setImageResource(iconResource);
            _currentIconResource = iconResource;
        }

        // update tag if necessary
        if (!_tag.getText().equals(tagText)) {
            _tag.setText(tagText);
        }
    }

    /**
     * A remote pool has been downloaded. Remove the progress bar, and change
     * the underlying item we're bound to to the new local pool.
     */
    public void onRemotePoolDownloadComplete(RemotePool pool,
                                             LocalPool localPool) {
        updateUI();
    }

    /**
     * There was an error downloading the remote pool. Switch back to line 2,
     * but change the text to show an error.
     */
    public void onRemotePoolDownloadError(RemotePool pool, String message) {
        updateUI();
    }

    /**
     * We've got a progress update on a download. Update the progress bar,
     * switch to it from line 2 if necessary.
     */
    public void onRemotePoolDownloadProgressUpdate(RemotePool pool, int progress) {
        updateUI();
    }

    /**
     * Set the local pools icon based on status.
     */
    public void onLocalPoolStatusChanged(LocalPool pool, Status oldStatus) {
        updateUI();
    }
}
