/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve;

import android.content.Context;
import android.content.SharedPreferences;

import androidx.preference.PreferenceManager;

/**
 * Preferences activity. Also includes accessors for our preferences, and
 * setting up the basic default preferences.
 */
public final class Preferences {
    static public final Object[] PREFS_LOCK = new Object[]{};
    private static final String KEY_DS_AUTOUPDATE = "dataStorageQueryForUpdates";
    private static final String KEY_TT_WRONG_ANSWER = "testTakingWrongAnswer";
    private static final String KEY_TT_RANDOM_ORDER = "testTakingRandomOrder";
    private static final String KEY_TT_CORRECT_RETRIES_STILL_WRONG = "testTakingCorrectRetryStillWrong";
    private static final String KEY_QB_MASK_ANSWERS = "questionBrowserMaskAnswers";
    private static final String KEY_QB_HIDE_WITHDRAWN = "questionBrowserHideWithdrawn";
    private static final boolean DEFAULT_DS_AUTOUPDATE = true;
    private static final String DEFAULT_TT_WRONG_ANSWER = "GONEXT";
    private static final boolean DEFAULT_TT_CORRECT_RETRIES_STILL_WRONG = false;
    private static final boolean DEFAULT_TT_RANDOM_ORDER = false;
    private static final boolean DEFAULT_QB_MASK_ANSWERS = true;
    private static final boolean DEFAULT_QB_HIDE_WITHDRAWN = true;
    static public boolean gEnsuredDefaults = false;

    /**
     * Make sure we have reasonable default values. Write default values to the
     * shared preferences, if necessary. This isn't needed to call the methods
     * to access our shared prefs - they will return sane defaults. It is
     * needed, though, to display the preferences UI, which otherwise will not
     * have the defaults set into it.
     * <p>
     * I could put the defaults in the XML... If I do that though, would they
     * change if I change the default of a preference and someone hadn't
     * overwrote it? I'm not sure. I will leave it this way for now.
     */
    static public void ensureDefaults(SharedPreferences prefs) {
        synchronized (PREFS_LOCK) {
            if (gEnsuredDefaults) {
                return;
            }

            boolean hasDsAutoUpdate = prefs.contains(KEY_DS_AUTOUPDATE);
            boolean hasTtWrongAnswer = prefs.contains(KEY_TT_WRONG_ANSWER);
            boolean hasTtCorrectRetriesStillWrong = prefs
                    .contains(KEY_TT_CORRECT_RETRIES_STILL_WRONG);
            boolean hasTtRandomOrder = prefs.contains(KEY_TT_RANDOM_ORDER);
            boolean hasQbMaskAnswers = prefs.contains(KEY_QB_MASK_ANSWERS);
            boolean hasQbHideWithdrawn = prefs.contains(KEY_QB_HIDE_WITHDRAWN);
            boolean missingPrefs = (!hasDsAutoUpdate
                    || !hasTtWrongAnswer || !hasQbMaskAnswers || !hasQbHideWithdrawn);

            if (missingPrefs) {
                SharedPreferences.Editor editor = prefs.edit();

                if (!hasDsAutoUpdate) {
                    editor.putBoolean(KEY_DS_AUTOUPDATE, DEFAULT_DS_AUTOUPDATE);
                }
                if (!hasTtWrongAnswer) {
                    editor.putString(KEY_TT_WRONG_ANSWER, DEFAULT_TT_WRONG_ANSWER);
                }
                if (!hasTtCorrectRetriesStillWrong) {
                    editor.putBoolean(KEY_TT_CORRECT_RETRIES_STILL_WRONG,
                            DEFAULT_TT_CORRECT_RETRIES_STILL_WRONG);
                }
                if (!hasTtRandomOrder) {
                    editor.putBoolean(KEY_TT_RANDOM_ORDER, DEFAULT_TT_RANDOM_ORDER);
                }
                if (!hasQbMaskAnswers) {
                    editor.putBoolean(KEY_QB_MASK_ANSWERS, DEFAULT_QB_MASK_ANSWERS);
                }
                if (!hasQbHideWithdrawn) {
                    editor.putBoolean(KEY_QB_HIDE_WITHDRAWN, DEFAULT_QB_HIDE_WITHDRAWN);
                }

                editor.apply();
                gEnsuredDefaults = true;
            }
        }
    }

    /**
     * Automatically check for updates?
     */
    static public boolean getAutomaticallyCheckForUpdates(Context context) {
        synchronized (PREFS_LOCK) {
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(context);
            ensureDefaults(prefs);
            return prefs.getBoolean(KEY_DS_AUTOUPDATE, DEFAULT_DS_AUTOUPDATE);
        }
    }

    /**
     * What to do when you answer questions incorrectly?
     */
    static public WrongAnswerBehavior getTestTakingWrongAnswer(Context context) {
        synchronized (PREFS_LOCK) {
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(context);
            ensureDefaults(prefs);
            String stringLocation = prefs.getString(KEY_TT_WRONG_ANSWER, DEFAULT_TT_WRONG_ANSWER);
            return WrongAnswerBehavior.valueOf(stringLocation);
        }
    }

    /**
     * When you retry a question you got a wrong answer on, should we still
     * consider it wrong for stats purposes?
     */
    static public boolean getCorrectRetriesStillWrong(Context context) {
        synchronized (PREFS_LOCK) {
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(context);
            ensureDefaults(prefs);
            return prefs.getBoolean(KEY_TT_CORRECT_RETRIES_STILL_WRONG,
                    DEFAULT_TT_CORRECT_RETRIES_STILL_WRONG);
        }
    }

    /**
     * Should questions be presented in random order?
     */
    static public boolean getTestTakingRandomOrder(Context context) {
        synchronized (PREFS_LOCK) {
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(context);
            ensureDefaults(prefs);
            return prefs.getBoolean(KEY_TT_RANDOM_ORDER, DEFAULT_TT_RANDOM_ORDER);
        }
    }

    /**
     * Mask answers in the question browser?
     */
    static public boolean getQuestionBrowserMaskAnswers(Context context) {
        synchronized (PREFS_LOCK) {
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(context);
            ensureDefaults(prefs);
            return prefs.getBoolean(KEY_QB_MASK_ANSWERS, DEFAULT_QB_MASK_ANSWERS);
        }
    }

    /**
     * Show withdrawn questions in the browser?
     */
    static public boolean getQuestionBrowserHideWithdrawn(Context context) {
        synchronized (PREFS_LOCK) {
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(context);
            ensureDefaults(prefs);
            return prefs.getBoolean(KEY_QB_HIDE_WITHDRAWN, DEFAULT_QB_HIDE_WITHDRAWN);
        }
    }

    @SuppressWarnings({"unused", "RedundantSuppression"})
    public enum WrongAnswerBehavior {
        PROMPT, GONEXT
    }
}
