/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve.activities;

import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.os.Bundle;
import android.view.ContextMenu;
import android.view.ContextMenu.ContextMenuInfo;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.widget.AdapterView;
import android.widget.AdapterView.AdapterContextMenuInfo;
import android.widget.AdapterView.OnItemClickListener;
import android.widget.ListView;
import android.widget.ProgressBar;

import androidx.appcompat.app.AppCompatActivity;

import net.rten.palmve.AboutBox;
import net.rten.palmve.LocalPool;
import net.rten.palmve.Pool;
import net.rten.palmve.PoolListAdapter;
import net.rten.palmve.Preferences;
import net.rten.palmve.R;
import net.rten.palmve.RemotePool;
import net.rten.palmve.repository.PoolDirectoryObserver;
import net.rten.palmve.repository.PoolRepo;

import javax.inject.Inject;

import dagger.hilt.android.AndroidEntryPoint;

@AndroidEntryPoint
public class PoolManagerActivity extends AppCompatActivity
        implements OnItemClickListener {
    private static final long UPDATE_CHECK_INTERVAL = 24 * 60 * 60 * 1000; // 1day
    private final PoolRepoEvents _events = new PoolRepoEvents();
    @Inject
    protected PoolRepo _pools;
    private PoolListAdapter _dataAdapter;
    private ListView _poolListView;
    private ProgressBar _progressBar;
    private MenuItem _refreshButton;

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.pool_manager);
        _poolListView = findViewById(R.id.pool_manager_list);
        _progressBar = findViewById(R.id.poolDirectoryProgress);
        poolDirectoryReady();
    }

    private void poolDirectoryReady() {
        // we need a data adapter for the list view
        _dataAdapter = new PoolListAdapter(this, _pools);
        _poolListView.setAdapter(_dataAdapter);

        // context menu support
        registerForContextMenu(_poolListView);

        // listen for select events
        _poolListView.setOnItemClickListener(this);

        // let typing invoke search by default
        setDefaultKeyMode(DEFAULT_KEYS_SEARCH_LOCAL);

        // should we download the remote pool list? yes if we have no pool
        // info at all
        boolean downloadPoolData = (_pools.getRemotePools() == null ||
                _pools.getRemotePools().isEmpty());

        // is it time for a periodic update?
        if (!downloadPoolData
                && Preferences.getAutomaticallyCheckForUpdates(this)) {
            long age = _pools.getAgeOfRemotePoolData();
            if (age < 0 || age >= UPDATE_CHECK_INTERVAL) {
                downloadPoolData = true;
            }
        }

        if (downloadPoolData) {
            downloadRemotePoolInfo();
        }

        // fake a notification if we're already downloading datums
        _pools.addObserver(_events);
        if (_pools.isRemotePoolInfoBeingDownloaded()) {
            _events.onDownloadingRemotePoolDataBeginning(_pools);
        }
    }

    /**
     * Popup our options menu.
     */
    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        MenuInflater inflater = getMenuInflater();
        inflater.inflate(R.menu.pool_directory_options, menu);
        _refreshButton = menu.findItem(R.id.poolDirectoryOptionsRefresh);
        showTitleBarSpinner(false);
        return true;
    }

    /**
     * Toggle the enable flag for refresh based on downloading state.
     */
    @Override
    public boolean onPrepareOptionsMenu(Menu menu) {
        boolean disableRefresh = false;

        // inflate the menu. should we allow refreshing the remote pool list?
        if (_pools != null) {
            if (_pools.isRemotePoolInfoBeingDownloaded()) {
                disableRefresh = true;
            } else {
                if (_pools.getRemotePools() != null) {
                    for (RemotePool remotePool : _pools
                            .getRemotePools()) {
                        if (remotePool.getStatus() == RemotePool.Status.rpDownloading) {
                            disableRefresh = true;
                            break;
                        }
                    }
                }
            }
        }

        menu.findItem(R.id.poolDirectoryOptionsRefresh).setEnabled(
                !disableRefresh);
        return true;
    }

    /**
     * Handle option menu entries.
     */
    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        if (item.getItemId() == R.id.poolDirectoryOptionsRefresh) {
            downloadRemotePoolInfo();
            return true;
        } else if (item.getItemId() == R.id.genericOptionsSearch) {
            onSearchRequested();
            return true;
        } else if (item.getItemId() == R.id.genericOptionsPreferences) {
            startActivity(new Intent(this, PreferencesActivity.class));
            return true;
        } else if (item.getItemId() == R.id.genericOptionsAbout) {
            AboutBox.getAboutBox(this).show();
            return true;
        } else {
            return super.onOptionsItemSelected(item);
        }
    }

    /**
     * Popup the correct context menu.
     */
    @Override
    public void onCreateContextMenu(ContextMenu menu, View v,
                                    ContextMenuInfo menuInfo) {
        super.onCreateContextMenu(menu, v, menuInfo);

        AdapterContextMenuInfo adapterContextMenuInfo = (AdapterContextMenuInfo) menuInfo;
        MenuInflater inflater = getMenuInflater();

        PoolListAdapter.ItemWrapper itemWrapper =
                (PoolListAdapter.ItemWrapper) _poolListView.getItemAtPosition(
                        adapterContextMenuInfo.position);
        if (itemWrapper.getObject() instanceof Pool) {
            Pool pool = (Pool) itemWrapper.getObject();
            if (pool instanceof RemotePool) {
                inflater.inflate(R.menu.pool_directory_cm_remote_pool, menu);
            } else if (pool instanceof LocalPool) {
                inflater.inflate(R.menu.pool_directory_cm_local_pool, menu);
            }
        }
    }

    /**
     * A context menu item has been selected.
     */
    @Override
    public boolean onContextItemSelected(MenuItem item) {
        AdapterContextMenuInfo info = (AdapterContextMenuInfo) item.getMenuInfo();
        PoolListAdapter.ItemWrapper itemWrapper =
                (PoolListAdapter.ItemWrapper) _poolListView.getItemAtPosition(info.position);
        Pool pool = (Pool) itemWrapper.getObject();

        if (item.getItemId() == R.id.poolDirectoryLocalDelete) {
            final LocalPool localPool = (LocalPool) pool;
            new AlertDialog.Builder(this)
                    .setMessage(getText(R.string.LocalPoolManager_DeleteConfirm))
                    .setPositiveButton(R.string.Standard_Yes, (dialog, id) -> {
                        if (id == DialogInterface.BUTTON_POSITIVE) {
                            _pools.deleteLocalPool(localPool);
                        }
                    })
                    .show();
            return true;
        } else if (item.getItemId() == R.id.poolDirectoryLocalHistory) {
            LocalPool localPool = (LocalPool) pool;
            Intent intent = new Intent(this, TestHistoryStatsActivity.class);
            intent.putExtra(TestHistoryStatsActivity.INTENT_POOL_NAME, localPool.getName());
            intent.putExtra(TestHistoryStatsActivity.INTENT_POOL_DATE, localPool.getDateField());
            intent.putExtra(GenericListActivity.INTENT_POOL_ID, localPool.getLocalPoolId());
            startActivity(intent);
            return true;
        } else if (item.getItemId() == R.id.poolDirectoryLocalBrowse) {
            LocalPool localPool = (LocalPool) pool;
            Intent intent = new Intent(this, ElementActivity.class);
            intent.putExtra(GenericListActivity.INTENT_POOL_ID, localPool.getLocalPoolId());
            intent.putExtra(ElementActivity.INTENT_ELEMENT_ID, localPool.getElementNr());
            startActivity(intent);
            return true;
        } else if (item.getItemId() == R.id.poolDirectoryLocalBrowseFigures) {
            LocalPool localPool = (LocalPool) pool;
            Intent intent = new Intent(this, FiguresActivity.class);
            intent.putExtra(GenericListActivity.INTENT_POOL_ID, localPool.getLocalPoolId());
            startActivity(intent);
            return true;
        } else if (item.getItemId() == R.id.poolDirectoryRemotePoolDownload) {
            RemotePool remotePool = (RemotePool) pool;
            if (remotePool.getStatus() != RemotePool.Status.rpDownloading) {
                downloadRemotePool((RemotePool) pool);
            }
            return true;
        } else {
            return super.onContextItemSelected(item);
        }
    }

    /**
     * Remove our event observer to keep us from getting callbacks after we're
     * gone.
     */
    @Override
    protected void onStop() {
        if (_pools != null) {
            _pools.removeObserver(_events);
        }
        super.onStop();
    }

    /**
     * Clear the list adapter. The way I wrote the list adapter has a memory
     * leak unless we tear it all down manually.
     */
    @Override
    protected void onDestroy() {
        super.onDestroy();
        if (_dataAdapter != null && _poolListView != null) {
            _dataAdapter.close(_poolListView);
        }
    }

    /**
     * Download a remote pool. Handle the error condition when external storage
     * is not available.
     */
    private void downloadRemotePool(final RemotePool remotePool) {
        // don't allow downloading a remote pool while we're refreshing
        if (_pools.isRemotePoolInfoBeingDownloaded()) {
            AlertDialog.Builder b = new AlertDialog.Builder(this);
            DialogInterface.OnClickListener clickListener = (dialog, id) -> {
            };

            // show the dialog
            b.setMessage(this
                    .getText(R.string.PoolManager_RefreshingTryAgainLater));
            b.setPositiveButton(R.string.Standard_Yes, clickListener);
            b.setNegativeButton(R.string.Standard_No, clickListener);
            b.show();
            return;
        }

        // We used to check for mounted external storage, but we don't bother and just
        // allow the download to fail at this point.
        _pools.downloadRemotePool(this, remotePool);
    }

    /**
     * Request that the pool directory download remote pool info.
     */
    private void downloadRemotePoolInfo() {
        if (!_pools.isRemotePoolInfoBeingDownloaded()) {
            // ask for remote pools
            _pools.requestRemotePools(this);
        }

        // TODO: if we have no local pools, we really should mark up the UI somehow. we used
        // to show a modal for that, but ProgressDlg has been deprecated.
    }

    /**
     * Clicking a remote pool starts to download it.
     */
    public void onItemClick(AdapterView<?> arg0, View listItemView,
                            int position, long id) {
        PoolListAdapter.ItemWrapper item = (PoolListAdapter.ItemWrapper) _dataAdapter.getItem(position);
        if (item.getObjectType() == PoolListAdapter.ItemWrapper.TYPE_POOL) {
            Pool pool = (Pool) item.getObject();
            RemotePool remotePool = null;
            LocalPool localPool = null;

            // check for update status
            if (pool instanceof LocalPool) {
                localPool = (LocalPool) pool;
                if (localPool.getStatus() == LocalPool.Status.lpUpdateAvailable) {
                    remotePool = localPool.getAssociatedRemotePool();
                }
            } else if (pool instanceof RemotePool) {
                remotePool = (RemotePool) pool;
            }

            if (remotePool != null
                    && remotePool.getStatus() != RemotePool.Status.rpDownloading) {
                // downloading or updating
                downloadRemotePool(remotePool);
            } else if (localPool != null) {
                Intent intent = new Intent(this, TestTakingActivity.class);
                intent.putExtra(GenericListActivity.INTENT_POOL_ID, localPool.getLocalPoolId());
                startActivity(intent);
            }
        } else if (item.getObjectType() == PoolListAdapter.ItemWrapper.TYPE_ALERT) {
            // this is the same as a refresh
            downloadRemotePoolInfo();
        }
    }

    /**
     * Turn on or off the title bar progress bar with the text provided.
     */
    private void showTitleBarSpinner(boolean show) {
        if (_progressBar != null && _refreshButton != null) {
            _progressBar.setVisibility(show ? View.VISIBLE : View.GONE);
            _refreshButton.setEnabled(!show);
        }
    }

    /**
     * Handles events in the pool repository.
     */
    private class PoolRepoEvents implements PoolDirectoryObserver {
        /**
         * Show the user that remote pool data is downloading.
         */
        @Override
        public void onDownloadingRemotePoolDataBeginning(PoolRepo directory) {
            showTitleBarSpinner(true);
        }

        /**
         * Clear any UI representation of remote pools being downloaded.
         */
        @Override
        public void onDownloadingRemotePoolDataCompleted(PoolRepo directory) {
            showTitleBarSpinner(false);
        }

        /**
         * Tell the user about the failure loading pool data.
         */
        @Override
        public void onDownloadingRemotePoolDataFailed(PoolRepo directory,
                                                      String message) {
            showTitleBarSpinner(false);

            // make an alert dialog about the failure if we already have some
            // local pools - because otherwise the error won't be displayed to
            // the user in the list view
            if (_pools.getLocalPools() != null && !_pools.getLocalPools().isEmpty()) {
                final AlertDialog.Builder builder = new AlertDialog.Builder(
                        PoolManagerActivity.this);
                builder.setMessage(PoolManagerActivity.this.getText(
                        R.string.PoolManager_LoadingFailed) + message);
                builder.setPositiveButton(R.string.Standard_Close,
                        (dialog, id) -> dialog.dismiss());
                final AlertDialog dlg = builder.create();
                dlg.show();
            }
        }

        /**
         * Our list adapter takes care of this, so we do nothing.
         */
        @Override
        public void onPoolDirectoryChanged(PoolRepo directory) {
        }
    }
}
