package net.rten.palmve.data.test;

import androidx.room.Dao;
import androidx.room.Query;
import androidx.room.Transaction;

import net.rten.palmve.data.pool.Answer;
import net.rten.palmve.data.pool.Element;
import net.rten.palmve.data.pool.QuestionView;

import java.util.Calendar;
import java.util.List;

/**
 * This DAO contains methods related to manipulating the current active test. This includes
 * generating a new test, figuring out the next question in the active test, grading a complete
 * test, etc...
 * <p>
 * TestHistoryDao contains methods for manipulating general test history, including listing the
 * questions in an active test.
 */
@Dao
public abstract class ActiveTestDao {
    /**
     * Return a list of active test IDs.
     *
     * @return Any active test ID, a list of 0, 1, or more items.
     */
    @Query("SELECT _id FROM TestHistory WHERE state = 1 AND element = :elementName AND elementdatefield = :elementDateField")
    public abstract List<Long> getActiveTestIds(String elementName, String elementDateField);

    /**
     * If there is an active test, return its ID, or return null. If there is more than one active
     * test, this will return the first one we find in the database.
     */
    public Long getActiveTestOrNull(String elementName, String elementDateField) {
        List<Long> activeTests = getActiveTestIds(elementName, elementDateField);
        if (activeTests.size() > 0) {
            return activeTests.get(0);
        }
        return null;
    }

    /**
     * Creates the proper entries in TestHistory and TestQuestions for a new test.
     *
     * @param testHistory DAO object for manipulating test history tables.
     * @param element     The element object to get name, date, version info from.
     * @param questions   The list of questions to populate the test.
     * @return The ID of the new test in the TestHistory table.
     */
    @Transaction
    public long createTest(TestHistoryDao testHistory, Element element,
                           List<QuestionView> questions) {
        // We need a test ID before we can add questions.
        long testId = testHistory.insertTest(Calendar.getInstance(), element.name, element.version,
                element.dateField, questions.size());

        // And we add questions in reverse order because we need each question to link to the
        // next. So we start at the end and work backwards.
        Long lastQuestionId = null;
        for (int i = questions.size() - 1; i >= 0; --i) {
            QuestionView q = questions.get(i);
            lastQuestionId = testHistory.insertQuestion(testId, q.id, q.designation,
                    q.questionGroupId, q.rightAnswer, i, lastQuestionId);
        }

        // The last question inserted is the first to ask, once we have it, set the first question
        // ID for the test we created.
        //noinspection ConstantConditions
        testHistory.setTestFirstQuestion(testId, lastQuestionId);
        return testId;
    }

    /**
     * Update the provided answer in a test's history data.
     */
    @Query("UPDATE TestQuestions SET providedanswer = :answer WHERE testid = :testId AND _id = :questionId")
    public abstract void submitAnswer(long testId, long questionId, Answer.AnswerCode answer);

    /**
     * Submit the final grading information for a test.
     */
    @Query("UPDATE TestHistory SET finalpercentage = :testScorePercent, datecompleted = :dateCompleted, state = :testState WHERE _id = :id")
    public abstract void completeTest(long id, TestState testState, int testScorePercent, Calendar dateCompleted);
}
