package net.rten.palmve.data.test;

import androidx.lifecycle.LiveData;
import androidx.room.Dao;
import androidx.room.Query;

import java.util.Calendar;
import java.util.List;

@Dao
public abstract class StatsDao {
    /**
     * Many different queries may run to load test stats depending on arguments. This method
     * will dispatch to the right one.
     *
     * @param element        Text name of the element to limit tests to.
     * @param elementDate    Text date field for the element.
     * @param reportType     Whether the report should aggregate stats at the question or question
     *                       group level.
     * @param correctAnswers If true, make stats about those questions or groups answered correctly.
     *                       If false, we'll make stats about missed questions or groups.
     * @param secondsAgo     This can be zero to disable filtering.
     * @param limit          Can be <= 0 to mean "no limit". Otherwise, limits the number of
     *                       returned rows.
     * @return The proper list of aggregated stats rows.
     */
    public LiveData<List<StatsView>> getStats(
            String element, String elementDate, DetailReportType reportType, boolean correctAnswers,
            int secondsAgo, int limit) {
        Calendar newerThan = null;
        if (secondsAgo > 0) {
            newerThan = Calendar.getInstance();
            newerThan.add(Calendar.SECOND, -secondsAgo);
        }
        if (limit == 0) {
            limit = -1;  // Sqlite3 interprets a negative as unlimited.
        }

        switch (reportType) {
            case questionGroups:
                if (correctAnswers) {
                    if (newerThan == null) {
                        return getQuestionGroupStatsWithCorrectAnswers(element, elementDate, limit);
                    } else {
                        return getQuestionGroupStatsWithCorrectAnswersNewerThan(element, elementDate, newerThan, limit);
                    }
                } else {
                    if (newerThan == null) {
                        return getQuestionGroupStatsWithIncorrectAnswers(element, elementDate, limit);
                    } else {
                        return getQuestionGroupStatsWithIncorrectAnswersNewerThan(element, elementDate, newerThan, limit);
                    }
                }

            case questions:
                if (correctAnswers) {
                    if (newerThan == null) {
                        return getQuestionStatsWithCorrectAnswers(element, elementDate, limit);
                    } else {
                        return getQuestionStatsWithCorrectAnswersNewerThan(element, elementDate, newerThan, limit);
                    }
                } else {
                    if (newerThan == null) {
                        return getQuestionStatsWithIncorrectAnswers(element, elementDate, limit);
                    } else {
                        return getQuestionStatsWithIncorrectAnswersNewerThan(element, elementDate, newerThan, limit);
                    }
                }
        }
        throw new RuntimeException("invalid value for reportType");
    }

    @Query("SELECT * FROM QuestionGroupStatsView WHERE answersWereCorrect = 0 AND element = :element " +
            "AND elementDate = :elementDate " +
            "GROUP BY questionGroupId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionGroupStatsWithIncorrectAnswers(
            String element, String elementDate, int limit);

    @Query("SELECT * FROM QuestionGroupStatsView WHERE answersWereCorrect = 1 AND element = :element " +
            "AND elementDate = :elementDate " +
            "GROUP BY questionGroupId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionGroupStatsWithCorrectAnswers(
            String element, String elementDate, int limit);

    @Query("SELECT * FROM QuestionStatsView WHERE answersWereCorrect = 0 AND element = :element " +
            "AND elementDate = :elementDate " +
            "GROUP BY questionId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionStatsWithIncorrectAnswers(
            String element, String elementDate, int limit);

    @Query("SELECT * FROM QuestionStatsView WHERE answersWereCorrect = 1 AND element = :element " +
            "AND elementDate = :elementDate " +
            "GROUP BY questionId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionStatsWithCorrectAnswers(
            String element, String elementDate, int limit);

    @Query("SELECT * FROM QuestionGroupStatsView WHERE answersWereCorrect = 0 AND element = :element " +
            "AND elementDate = :elementDate AND testCompletionDate >= :newerThan " +
            "GROUP BY questionGroupId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionGroupStatsWithIncorrectAnswersNewerThan(
            String element, String elementDate, Calendar newerThan, int limit);

    @Query("SELECT * FROM QuestionGroupStatsView WHERE answersWereCorrect = 1 AND element = :element " +
            "AND elementDate = :elementDate AND testCompletionDate >= :newerThan " +
            "GROUP BY questionGroupId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionGroupStatsWithCorrectAnswersNewerThan(
            String element, String elementDate, Calendar newerThan, int limit);

    @Query("SELECT * FROM QuestionStatsView WHERE answersWereCorrect = 0 AND element = :element " +
            "AND elementDate = :elementDate AND testCompletionDate >= :newerThan " +
            "GROUP BY questionId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionStatsWithIncorrectAnswersNewerThan(
            String element, String elementDate, Calendar newerThan, int limit);

    @Query("SELECT * FROM QuestionStatsView WHERE answersWereCorrect = 1 AND element = :element " +
            "AND elementDate = :elementDate AND testCompletionDate >= :newerThan " +
            "GROUP BY questionId, designation ORDER BY questionCount DESC, designation ASC " +
            "LIMIT :limit")
    public abstract LiveData<List<StatsView>> getQuestionStatsWithCorrectAnswersNewerThan(
            String element, String elementDate, Calendar newerThan, int limit);

    public enum DetailReportType {
        questionGroups,  // Stats grouped by question group.
        questions        // Stats grouped by individual questions.
    }
}
