package net.rten.palmve.models;

import androidx.lifecycle.LiveData;
import androidx.lifecycle.MutableLiveData;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModel;

import net.rten.palmve.data.pool.PoolDatabase;
import net.rten.palmve.data.pool.QuestionView;
import net.rten.palmve.data.pool.QuestionViewForAnswerSearch;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class PoolSearchResultsModel extends ViewModel {
    // Our search string:
    private final MutableLiveData<String> _term = new MutableLiveData<>();

    // Live data from the searches we do over each DB:
    private final List<LiveData<List<QuestionView>>> _searchQ = new ArrayList<>();
    private final List<LiveData<List<QuestionViewForAnswerSearch>>> _searchA = new ArrayList<>();
    private final List<RememberAndMerge<QuestionView>> _copyResultsQ = new ArrayList<>();
    private final List<RememberAndMerge<QuestionViewForAnswerSearch>> _copyResultsA = new ArrayList<>();

    // Output from the model:
    private final MutableLiveData<List<QuestionView>> _results = new MutableLiveData<>();

    private class RememberAndMerge<V> implements Observer<List<V>> {
        private final int _highOrderIdBits;
        private List<V> _list;

        public RememberAndMerge(int highOrderIdBits) {
            _highOrderIdBits = highOrderIdBits;
        }

        public List<V> list() {
            return _list;
        }

        public int highOrderBits() {
            return _highOrderIdBits;
        }

        @Override
        public void onChanged(List<V> data) {
            _list = data;
            mergeResults();
        }
    }

    public void setSearchTerm(String term) {
        _term.setValue(term);
    }

    public LiveData<List<QuestionView>> getSearchResults() {
        return _results;
    }

    public void setUp(List<PoolDatabase> dbs, List<Integer> ids) {
        if (_copyResultsQ.isEmpty()) {
            if (dbs.size() != ids.size()) {
                throw new IllegalArgumentException("lists should have the same size");
            }
            for (int i = 0; i < dbs.size(); ++i) {
                PoolDatabase poolDb = dbs.get(i);
                Integer id = ids.get(i);

                LiveData<List<QuestionView>> sq = poolDb.questionDao().searchQuestionsLiveData(_term.getValue());
                LiveData<List<QuestionViewForAnswerSearch>> sa = poolDb.questionDao().searchQuestionAnswersLiveData(_term.getValue());
                RememberAndMerge<QuestionView> cq = new RememberAndMerge<>(id);
                RememberAndMerge<QuestionViewForAnswerSearch> ca = new RememberAndMerge<>(id);
                _copyResultsQ.add(cq);
                _copyResultsA.add(ca);
                sq.observeForever(cq);
                sa.observeForever(ca);
                _searchQ.add(sq);
                _searchA.add(sa);
            }
        }
    }

    @Override
    public void onCleared() {
        if (!_copyResultsQ.isEmpty()) {
            for (int i = 0; i < _searchQ.size(); ++i) {
                _searchQ.get(i).removeObserver(_copyResultsQ.get(i));
            }
            for (int i = 0; i < _searchA.size(); ++i) {
                _searchA.get(i).removeObserver(_copyResultsA.get(i));
            }
            _searchQ.clear();
            _searchA.clear();
            _copyResultsQ.clear();
            _copyResultsA.clear();
        }
    }

    private void mergeResults() {
        List<QuestionView> merged = new ArrayList<>();
        Set<Long> seenQIds = new HashSet<>();

        for (int i = 0; i < _copyResultsQ.size(); ++i) {
            // We need to merge results at the question and answer level.
            if (_copyResultsQ.get(i).list() != null) {
                for (int j = 0; j < _copyResultsQ.get(i).list().size(); ++j) {
                    QuestionView dupe = duplicateItem(j, _copyResultsQ.get(i));
                    seenQIds.add(dupe.id);
                    merged.add(dupe);
                }
            }

            // For answers, we actually have more info then we expose. We just flatten everything
            // down into a list of QuestionView, even though we know the answer text and letter that
            // matched the search term.
            if (_copyResultsA.get(i).list() != null) {
                for (int j = 0; j < _copyResultsA.get(i).list().size(); ++j) {
                    QuestionView dupe = duplicateItem(j, _copyResultsA.get(i));
                    if (!seenQIds.contains(dupe.id)) {
                        seenQIds.add(dupe.id);
                        merged.add(dupe);
                    }
                }
            }
        }

        _results.setValue(merged);
    }

    static private QuestionView duplicateItem(int index, RememberAndMerge<? extends QuestionView> r) {
        QuestionView original = r.list().get(index);
        QuestionView copy = new QuestionView();
        copy.description = original.description;
        copy.designation = original.designation;
        copy.id = ((long) r.highOrderBits() << 16) | original.id;
        return copy;
    }
}
