package net.rten.palmve.models;

import androidx.lifecycle.LiveData;
import androidx.lifecycle.MutableLiveData;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModel;

import net.rten.palmve.data.pool.PoolDatabase;
import net.rten.palmve.data.pool.QuestionView;
import net.rten.palmve.data.test.TestDatabase;
import net.rten.palmve.data.test.TestHistory;
import net.rten.palmve.data.test.TestQuestions;
import net.rten.palmve.utilities.CopyObserver;

import java.util.ArrayList;
import java.util.List;

public class TestSummaryModel extends ViewModel {
    private final CopyObserver<TestHistory> _copyTestInfo = new CopyObserver<>();
    private final CopyTestQuestions _copyTestQuestions = new CopyTestQuestions();
    private final MutableLiveData<Integer> _itemChanges = new MutableLiveData<>();
    private LiveData<List<TestQuestions>> _questions;
    private PoolDatabase _poolDb;
    private int _poolId;

    public int getPoolId() {
        return _poolId;
    }

    public LiveData<List<TestQuestionData>> getTestQuestions() {
        return _copyTestQuestions.getQuestionData();
    }

    public LiveData<TestHistory> getTestInfo() {
        return _copyTestInfo.getCopy();
    }

    public LiveData<Integer> getTestItemUpdated() {
        return _itemChanges;
    }

    public void setUp(int poolId, PoolDatabase poolDb, TestDatabase testDb, long testId) {
        if (_poolDb == null) {
            _poolDb = poolDb;
            _poolId = poolId;
            _copyTestInfo.startObservingMaster(testDb.testDao().getTestInfoByIdLiveData(testId));
            _questions = testDb.testDao().getQuestionsInTestLiveData(testId);
            _questions.observeForever(_copyTestQuestions);
        }
    }

    @Override
    public void onCleared() {
        _copyTestQuestions.clear();
        _copyTestInfo.stopObservingMaster();
        if (_questions != null) {
            _questions.removeObserver(_copyTestQuestions);
            _questions = null;
        }
    }

    /**
     * Manages the state of loading the data from the pool DB once we have the question list
     * from the test DB.
     */
    private static class Resolver {
        public LiveData<QuestionView> question;
        public Observer<QuestionView> observer;
    }

    /**
     * Observes the test data stored in the test history DB, and fires off queries to the pool
     * database to load the actual question contents (description, designation, etc...).
     */
    private class CopyTestQuestions implements Observer<List<TestQuestions>> {
        private final MutableLiveData<List<TestQuestionData>> _questionData = new MutableLiveData<>();
        final private ArrayList<Resolver> _resolvers = new ArrayList<>();

        public LiveData<List<TestQuestionData>> getQuestionData() {
            return _questionData;
        }

        public void clear() {
            for (Resolver r : _resolvers) {
                r.question.removeObserver(r.observer);
            }
            _resolvers.clear();
        }

        @Override
        public void onChanged(List<TestQuestions> item) {
            clear();
            ArrayList<TestQuestionData> data = new ArrayList<>();
            for (int i = 0; i < item.size(); ++i) {
                // Create our class that combines both pieces of data.
                final TestQuestionData copy = new TestQuestionData();
                copy.testQuestion = item.get(i);
                data.add(copy);

                // Now make the resolver object, which wraps our query to get the actual question
                // data from the pool DB.
                Resolver r = new Resolver();
                r.question = _poolDb.questionDao().getQuestionByIdLiveData(
                        copy.testQuestion.questionId);
                final int finalIdx = i;
                r.observer = question -> {
                    copy.poolQuestion = question;
                    _itemChanges.setValue(finalIdx);
                };
                r.question.observeForever(r.observer);
                _resolvers.add(r);
            }

            // TODO: the old app used to sort by correct and then designator. that's fine, but we
            // can't simply sort the list because we need to manage _itemChanges properly when we
            // get the descriptor string.

            _questionData.setValue(data);
        }
    }
}
