package net.rten.palmve.models;

import android.util.Log;

import androidx.annotation.NonNull;
import androidx.lifecycle.LiveData;
import androidx.lifecycle.MutableLiveData;
import androidx.lifecycle.ViewModel;

import net.rten.palmve.data.pool.Answer;
import net.rten.palmve.data.pool.PoolDatabase;
import net.rten.palmve.data.pool.QuestionView;
import net.rten.palmve.data.test.TestDatabase;
import net.rten.palmve.data.test.TestHistory;
import net.rten.palmve.data.test.TestQuestions;
import net.rten.palmve.data.test.TestState;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.List;

/**
 * The test taking model is responsible for traking the state of an active test. This includes
 * creating an active test if one is not already in progress.
 */
public class TestTakingModel extends ViewModel {

    private final static String TAG = "TestTakingModel";

    /**
     * I always thought this was 75, but looked it up and it's 74. 37/50 on extra, or 26/35 on
     * tech or general.
     */
    final static private int PASSING_PERCENTAGE = 74;

    final private MutableLiveData<Long> _testId = new MutableLiveData<>();
    final private MutableLiveData<TestHistory> _testInfo = new MutableLiveData<>();
    final private MutableLiveData<List<TestQuestionData>> _questions = new MutableLiveData<>();
    final private MutableLiveData<Integer> _currentQuestionIndex = new MutableLiveData<>();
    final private MutableLiveData<Integer> _lastQuestionChangeIndex = new MutableLiveData<>();
    final private MutableLiveData<TestProgress> _testProgress = new MutableLiveData<>();
    private PoolDatabase _poolDb;
    private TestDatabase _testDb;
    private String _elementName;
    private String _elementDateField;
    private boolean _correctRetriesStillWrong;
    private boolean _randomizeQuestionOrder = false;

    public LiveData<Long> getTestId() {
        return _testId;
    }

    public LiveData<TestHistory> getTestInfo() {
        return _testInfo;
    }

    public LiveData<List<TestQuestionData>> getTestQuestions() {
        return _questions;
    }

    public LiveData<Integer> getCurrentQuestionIndex() {
        return _currentQuestionIndex;
    }

    public LiveData<TestProgress> getTestProgress() {
        return _testProgress;
    }

    public boolean getCorrectAnswersStillWrong() {
        return _correctRetriesStillWrong;
    }

    public void setCorrectAnswersStillWrong(boolean correctRetriesStillWrong) {
        _correctRetriesStillWrong = correctRetriesStillWrong;
    }

    public void setRandomizeQuestionOrder(boolean to) {
        _randomizeQuestionOrder = to;
    }

    /**
     * Load the current test state, or create a new test if one is not active. Initializes our
     * observables. This process happens asynchronously in another thread.
     */
    public void setUp(String elementName, String elementDateField, PoolDatabase poolDb, TestDatabase testDb) {
        if (_poolDb == null) {
            _poolDb = poolDb;
            _testDb = testDb;
            _elementName = elementName;
            _elementDateField = elementDateField;

            // We'll do the initial loading and setup in a thread, since it's way easier than
            // trying to deal with the sort of asynchronous API with LiveData. However, I really
            // don't know what to do with something failing in this path.
            new Thread(this::loadOrCreateTest).start();
        }
    }

    /**
     * Helper to submit an answer to the active question. This version of the method will also
     * advance the current question index, if we are not already at the last question in the test.
     */
    public void submitAnswerToActiveQuestion(Answer.AnswerCode answer) {
        if (_currentQuestionIndex.getValue() == null || _questions.getValue() == null) {
            Log.e(TAG, "during answer submission, there is no pending question");
            return;
        }
        Integer currentIndex = _currentQuestionIndex.getValue();
        submitAnswer(currentIndex++, answer);
        if (currentIndex >= _questions.getValue().size()) {
            currentIndex = null;
        }
        _currentQuestionIndex.postValue(currentIndex);
    }

    /**
     * Set the answer for one of the
     */
    public void submitAnswer(int questionIndex, Answer.AnswerCode answer) {
        List<TestQuestionData> questions = _questions.getValue();
        Long testId = _testId.getValue();
        if (questions == null || testId == null) {
            throw new IllegalStateException("data is not loaded yet");
        }
        if (questionIndex < 0 || questionIndex >= questions.size()) {
            throw new IllegalArgumentException("question index out of bounds");
        }
        TestQuestionData questionData = questions.get(questionIndex);
        _testDb.activeTestDao().submitAnswer(testId, questionData.testQuestion.id, answer);

        // We need to just override the data in 'testQuestion', since there's no LiveData connection
        // that the update would trigger.
        questionData.testQuestion.providedAnswer = answer;
        _lastQuestionChangeIndex.postValue(questionIndex);
        updateTestProgress(questions);
    }

    /**
     * Move the test to the finished state, and write the final grade and timestamp to the database.
     * Requires that the test be complete (_testProgress.allQuestionsAnswered is true).
     */
    public void completeTest() {
        TestProgress progress = _testProgress.getValue();
        TestHistory info = _testInfo.getValue();
        if (progress == null || info == null) {
            throw new IllegalStateException("data not loaded yet");
        }
        if (!progress.allQuestionsAnswered) {
            throw new IllegalStateException("all questions not yet answered");
        }
        info.testScorePercent = progress.percentCorrect;
        info.testState = info.testScorePercent >= PASSING_PERCENTAGE ? TestState.PASSED : TestState.FAILED;
        info.dateCompleted = Calendar.getInstance();

        _testDb.activeTestDao().completeTest(info.id, info.testState, info.testScorePercent,
                info.dateCompleted);
        _testInfo.postValue(info);
    }

    /**
     * Manages the process of loading or creating a new test. Designed to be called from something
     * other than the main thread.
     */
    private void loadOrCreateTest() {
        List<QuestionView> questionViews = null;
        Long currentTestId = _testDb.activeTestDao().getActiveTestOrNull(
                _elementName, _elementDateField);
        if (currentTestId == null) {
            // We need to make a new test, which entails generating a question list.
            questionViews = _poolDb.testGenerationDao().generateTest();
            if (_randomizeQuestionOrder) {
                Collections.shuffle(questionViews);
            }
            currentTestId = _testDb.activeTestDao().createTest(_testDb.testDao(),
                    _poolDb.elementDao().getElement(), questionViews);
        }

        // Now we can load the question history piece, the rows for each question. This is
        // not very efficient, since during test creation we already had this info.
        List<TestQuestions> questionHistories = _testDb.testDao().getQuestionsInTest(currentTestId);
        if (questionViews == null) {
            // If we didn't create the test, we don't have the question information, yet.
            questionViews = new ArrayList<>();
            for (TestQuestions q : questionHistories) {
                questionViews.add(_poolDb.questionDao().getQuestionById(q.questionId));
            }
        }

        // Now we can combine the data from the two sources.
        ArrayList<TestQuestionData> questionData = new ArrayList<>();
        int currentQuestion = -1;
        for (int i = 0; i < questionViews.size(); ++i) {
            TestQuestionData testQuestionData = new TestQuestionData();
            testQuestionData.poolQuestion = questionViews.get(i);
            testQuestionData.testQuestion = questionHistories.get(i);
            questionData.add(testQuestionData);

            // If this was the first question we've found without an answer, it's the active
            // question in the test.
            if (currentQuestion < 0 && testQuestionData.testQuestion.providedAnswer == null) {
                currentQuestion = i;
            }
        }

        // Now load the TestHistory object - again, inefficient if we created the test.
        TestHistory testInfo = _testDb.testDao().getTestInfoById(currentTestId);

        // Post the data we've loaded to our observables.
        _testId.postValue(currentTestId);
        _testInfo.postValue(testInfo);
        _questions.postValue(questionData);
        _currentQuestionIndex.postValue(currentQuestion);
        updateTestProgress(questionData);
    }

    /**
     * Updates _testProgress with current progress information.
     */
    private void updateTestProgress(@NonNull List<TestQuestionData> questions) {
        TestProgress progress = new TestProgress();
        for (TestQuestionData data : questions) {
            if (data.testQuestion.providedAnswer != null) {
                ++progress.answersProvided;
                if (data.testQuestion.providedAnswer.equals(data.testQuestion.correctAnswer)) {
                    ++progress.answersCorrect;
                }
            }
        }
        if (progress.answersProvided != 0) {
            progress.percentAnswered = (progress.answersProvided * 100) / questions.size();
            progress.percentCorrect = (progress.answersCorrect * 100) / progress.answersProvided;
        }
        progress.allQuestionsAnswered = (progress.answersProvided == questions.size());
        _testProgress.postValue(progress);
    }

    /**
     * Provides a view of the progress for the current test.
     */
    public static class TestProgress {
        public int answersProvided = 0;
        public int percentAnswered = 0;
        public int answersCorrect = 0;
        public int percentCorrect = 0;
        public boolean allQuestionsAnswered = false;
    }
}
