/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve.repository;

import android.app.Activity;
import android.content.Context;
import android.util.Log;
import android.util.Xml;

import androidx.annotation.VisibleForTesting;

import net.rten.palmve.LocalPool;
import net.rten.palmve.RemotePool;
import net.rten.palmve.asynctasks.FetchRemotePoolAsyncTask;
import net.rten.palmve.asynctasks.FetchRemotePoolListAsyncTask;
import net.rten.palmve.utilities.RemotePoolXmlParser;

import java.io.File;
import java.io.FileInputStream;
import java.net.URL;
import java.util.ArrayList;

import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import dagger.hilt.InstallIn;
import dagger.hilt.android.qualifiers.ApplicationContext;
import dagger.hilt.components.SingletonComponent;

/**
 * This class manages access to all the question pools either on device or
 * available for download.
 *
 * @author Michael Ballbach
 */
@Module
@InstallIn(SingletonComponent.class)
public class PoolRepo {
    // this lock is surfaced, and must be held to call various DB interacting
    // methods on local pools, along with the content provider using it to
    // keep us from closing the database out from under them
    public static final Object[] POOL_LOCK = new Object[0];
    private static final String TAG = "PoolDirectory";
    private static final String POOLS_URL = "https://palmve.net/android-pools/";
    private static final String POOL_XML_URL = POOLS_URL + "pools.xml";
    private static final String DB_EXT = ".sqlite3";
    private static final String DOWNLOAD_EXT_TAG = "-downloading";
    private static final String DOWNLOAD_EXT = DB_EXT + DOWNLOAD_EXT_TAG;

    private static final long CACHED_REMOTE_POOL_AGE_LIMIT = 30 * 24 * 60
            * 60 * 1000L; // one month or so
    private final ArrayList<LocalPool> _localPools = new ArrayList<>();
    private final ArrayList<RemotePool> _remotePools = new ArrayList<>();
    private final ArrayList<PoolDirectoryObserver> _eventObservers = new ArrayList<>();
    private final Context _context;
    private final String _dataDir;
    private final File _cachedRemotePoolListFile;
    private int _localPoolId;
    private boolean _fetchingRemotePools;

    /**
     * The constructor allows for specifying which directories to use for files.
     *
     * @param context An app-level or equivalent context.
     * @param dataDir Where on the filesystem we should look for and store question pool DBs.
     */
    public PoolRepo(Context context, String dataDir) {
        _context = context;
        _dataDir = dataDir + File.separator;
        _cachedRemotePoolListFile = new File(_dataDir + "cached_remote_pools.xml");

        // Start a background thread to look for actual pools.
        new Thread(() -> {
            attachLocalPools();
            loadRemotePoolCache();
            synchronized (_localPools) {
                _localPools.notifyAll();
            }
        }).start();
    }

    @Singleton
    @Provides
    public static PoolRepo providePoolRepo(@ApplicationContext Context appContext) {
        final String dataBaseDirectory = appContext.getFilesDir().getParent();
        return new PoolRepo(appContext, dataBaseDirectory + File.separator + "databases");
    }

    /**
     * Exposed primarily for testing, to allow the asynchronous pool setup to be waited on.
     */
    @VisibleForTesting
    public void waitForLocalPools() {
        synchronized (_localPools) {
            while (_localPools.size() <= 0) {
                try {
                    _localPools.wait();
                } catch (InterruptedException exc) {
                    // Ignore
                }
            }
        }
    }

    /**
     * Look for local pools. Create objects to represent them and place them in
     * the local pools directory. Clear any existing local pools.
     */
    private void attachLocalPools() {
        // clear out anything currently in there
        for (LocalPool localPool : _localPools) {
            localPool.close();
        }
        _localPools.clear();

        synchronized (POOL_LOCK) {
            // now look for pool files
            final File probeDirectory = new File(_dataDir);

            if (!probeDirectory.exists()) {
                if (!probeDirectory.mkdir()) {
                    throw new RuntimeException("failed to create directory " + probeDirectory.toString());
                }
            }

            String[] entries = probeDirectory.list();
            if (entries != null) {
                for (String filename : entries) {
                    File filepath = new File(probeDirectory, filename);
                    if (filename.endsWith(DB_EXT)) {
                        try {
                            // load the local pool
                            LocalPool pool = new LocalPool(_localPoolId++);
                            pool.bindToFile(_context, filepath);
                            _localPools.add(pool);
                        } catch (Exception e) {
                            // not sure what to do here. just delete it.
                            if (!filepath.delete()) {
                                Log.w(TAG, "failed to delete " + filepath.toString());
                            }
                        }
                    } else if (filename.endsWith(DOWNLOAD_EXT)) {
                        // we've got an old in-progress download file,
                        // delete it
                        if (!filepath.delete()) {
                            Log.w(TAG, "failed to delete " + filepath.toString());
                        }
                    }
                }
            }
        }
    }

    public void addObserver(PoolDirectoryObserver observer) {
        _eventObservers.add(observer);
    }

    public void removeObserver(PoolDirectoryObserver observer) {
        _eventObservers.remove(observer);
    }

    public ArrayList<LocalPool> getLocalPools() {
        return _localPools;
    }

    /**
     * Find a local pool by ID.
     *
     * @param id The local pool's runtime ID. Not the element number.
     * @return The pool, if found.
     */
    public LocalPool getLocalPoolById(int id) {
        for (LocalPool localPool : _localPools) {
            if (localPool.getLocalPoolId() == id) {
                return localPool;
            }
        }
        return null;
    }

    /**
     * Returns the list of remote pools, or sometimes null if there aren't any.
     */
    public ArrayList<RemotePool> getRemotePools() {
        return _remotePools;
    }

    /**
     * Find a local pool by name and datefield. Returns null if it can't be
     * found.
     */
    public LocalPool findLocalPool(String name, String dateField) {
        for (LocalPool localPool : _localPools) {
            if (localPool.getName().equals(name)
                    && localPool.getDateField().equals(dateField)) {
                return localPool;
            }
        }
        return null;
    }

    /**
     * Request that remote pools get downloaded. This will happen
     * asynchronously. Once they're downloaded, our observers will be alerted
     * with the pool changes.
     *
     * @param activity We'll call interface callbacks on activity's UI thread.
     */
    public void requestRemotePools(Activity activity) {
        synchronized (POOL_LOCK) {
            if (_fetchingRemotePools) {
                return;
            }
            _fetchingRemotePools = true;
        }

        new Thread(() -> {
            try {
                // let observers know we've started to fetch this info
                activity.runOnUiThread(() -> {
                    for (PoolDirectoryObserver observer : _eventObservers) {
                        observer.onDownloadingRemotePoolDataBeginning(this);
                    }
                });

                final URL url = new URL(POOL_XML_URL);
                ArrayList<RemotePool> pools = FetchRemotePoolListAsyncTask.run(
                        url, _cachedRemotePoolListFile);
                setRemotePoolInfo(pools);

                activity.runOnUiThread(() -> {
                    alertObservers();
                    for (PoolDirectoryObserver observer : _eventObservers) {
                        observer.onDownloadingRemotePoolDataCompleted(this);
                    }
                });
            } catch (Exception e) {
                final String error = e.getClass().getName() + ": " + e.getMessage();
                activity.runOnUiThread(() -> {
                    for (PoolDirectoryObserver observer : _eventObservers) {
                        observer.onDownloadingRemotePoolDataFailed(this, error);
                    }
                });
            } finally {
                synchronized (POOL_LOCK) {
                    _fetchingRemotePools = false;
                }
            }
        }).start();
    }

    /**
     * Is remote pool data being downloaded?
     *
     * @return True if so.
     */
    public boolean isRemotePoolInfoBeingDownloaded() {
        return _fetchingRemotePools;
    }

    /**
     * Request downloading a remote pool. This will begin to download it to the
     * local device.
     *
     * @param activity Callbacks will run in activity's UI thread
     * @param pool     The pool that should be downloaded
     */
    public void downloadRemotePool(Activity activity, RemotePool pool) {
        new Thread(() -> {
            try {
                final long[] lastStatusUpdate = {0};
                URL poolURL = new URL(POOLS_URL + pool.getRemoteFilename());
                final LocalPool upgradesPool = pool.getAssociatedLocalPool();
                final File directoryToSaveTo = new File(_dataDir);

                // tell the remote pool what it's file name will be, start the task
                File destinationFile = new File(directoryToSaveTo, pool.getRemoteFilename());
                if (upgradesPool != null) {
                    synchronized (POOL_LOCK) {
                        upgradesPool.close();
                    }
                }

                // Do the downloading.
                FetchRemotePoolAsyncTask.run(poolURL, destinationFile, pool.getSum(),
                        (int readSoFar, int contentLength) -> {
                            long now = System.currentTimeMillis();
                            if (now - lastStatusUpdate[0] > 250) {
                                final int progress =
                                        contentLength > 0 ? readSoFar * 100 / contentLength : -1;
                                activity.runOnUiThread(() -> pool.setDownloading(progress));
                                lastStatusUpdate[0] = now;
                            }
                        });

                // bind the new file into a LocalPool and swap it into the remote pool
                synchronized (POOL_LOCK) {
                    LocalPool localPool = new LocalPool(_localPoolId++);
                    try {
                        localPool.bindToFile(_context, destinationFile);
                        localPool.setAssociatedRemotePool(pool);
                        pool.setAssociatedLocalPool(localPool);
                        _localPools.add(localPool);
                    } catch (Exception e) {
                        // If something in that process went wrong, try to put the old local pool
                        // back, delete the downloaded file, and re-throw to propagate the error
                        // message.
                        if (!destinationFile.delete()) {
                            Log.w(TAG, "failed to delete " + destinationFile.toString());
                        }
                        pool.setAssociatedLocalPool(upgradesPool);
                        _localPools.remove(localPool);
                        throw e;
                    }
                }

                // now we can post the download state change to update the UI
                activity.runOnUiThread(() -> {
                    pool.setDownloaded();
                    alertObservers();
                });
            } catch (Exception e) {
                activity.runOnUiThread(() -> pool.setDownloadFailed(e.getMessage()));
            }
        }).start();
    }

    /**
     * Delete a local pool.
     */
    public void deleteLocalPool(LocalPool pool) {
        File filename = pool.getUnderlyingFilename();

        synchronized (POOL_LOCK) {
            // clear links, close and delete the pool
            if (pool.getAssociatedRemotePool() != null) {
                pool.getAssociatedRemotePool().setAssociatedLocalPool(null);
            }
            pool.close();
            _localPools.remove(pool);
            if (filename != null) {
                if (!filename.delete()) {
                    throw new RuntimeException("failed to delete file " + filename.toString());
                }
            }
        }

        // let our listeners know the pool list may have changed
        alertObservers();
    }

    /**
     * How old is the remote pool data? This returns the timestamp of the remote
     * pool list cache file.
     *
     * @return An age in milliseconds. This can return -1 if there is no valid
     * cached remote pool data or the age appears to be invalid.
     */
    public long getAgeOfRemotePoolData() {
        long returnValue = -1;
        if (_cachedRemotePoolListFile.exists()) {
            returnValue = System.currentTimeMillis()
                    - _cachedRemotePoolListFile.lastModified();
            if (returnValue < 0) {
                returnValue = -1;
            }
        }
        return returnValue;
    }

    /**
     * Call any registered event handler about the directory list changing.
     */
    private void alertObservers() {
        // alert observers
        for (PoolDirectoryObserver observer : _eventObservers) {
            observer.onPoolDirectoryChanged(this);
        }
    }

    /**
     * Set our remote pool list, and link remote pools to their local
     * counterparts, and vice-versa. If called with 'null', the function will
     * attempt to link existing local and remote pools, but not reset the list
     * of remote pools.
     */
    private void setRemotePoolInfo(ArrayList<RemotePool> pools) {
        synchronized (POOL_LOCK) {
            // clear any current local -> remote pool associations
            for (LocalPool localPool : _localPools) {
                localPool.setAssociatedRemotePool(null);
            }

            // store the remote pools, iterate them and see if any are updates for
            // our local pools. if so, associate them.
            if (pools != null) {
                _remotePools.clear();
                _remotePools.addAll(pools);
            }
            for (RemotePool remotePool : _remotePools) {
                for (LocalPool localPool : _localPools) {
                    if (remotePool.getName().equals(localPool.getName())
                            && remotePool.getDateField().equals(
                            localPool.getDateField())) {
                        remotePool.setAssociatedLocalPool(localPool);
                        localPool.setAssociatedRemotePool(remotePool);
                        break;
                    }
                }
            }
        }
    }

    /**
     * If we have any cached remote pool data, load it.
     */
    private void loadRemotePoolCache() {
        if (_cachedRemotePoolListFile.exists()) {
            long lastModified = _cachedRemotePoolListFile.lastModified();
            long age = System.currentTimeMillis() - lastModified;

            if (age >= 0 && age < CACHED_REMOTE_POOL_AGE_LIMIT) {
                try {
                    FileInputStream inputStream = new FileInputStream(_cachedRemotePoolListFile);
                    RemotePoolXmlParser parser = new RemotePoolXmlParser();
                    Xml.parse(inputStream, Xml.Encoding.US_ASCII, parser);
                    setRemotePoolInfo(parser.getRemotePools());
                } catch (Exception e) {
                    // don't notify the user or anything. just the cache is bad.
                    if (!_cachedRemotePoolListFile.delete()) {
                        throw new RuntimeException("failed to delete file " +
                                _cachedRemotePoolListFile.toString());
                    }
                }
            }
        }
    }
}
